from typing import Any, Dict, Optional

from ape.api import ConverterAPI
from ape.exceptions import NetworkError, ProviderError
from ape.logging import logger
from ape.types import AddressType
from ape.utils import cached_property
from ape_ethereum.provider import Web3Provider
from web3.exceptions import CannotHandleRequest
from web3.main import ENS


class ENSConversions(ConverterAPI):
    """Converts ENS names like `my-name.eth` to `0xAbCd...1234`"""

    address_cache: Dict[str, AddressType] = {}

    @cached_property
    def mainnet_provider(self) -> Optional[Web3Provider]:
        provider = self.network_manager.active_provider
        if (
            provider
            and isinstance(provider, Web3Provider)
            and provider.network.name in ("mainnet", "mainnet-fork")
            and provider.network.ecosystem.name == "ethereum"
        ):
            return provider

        # Connect to mainnet for ENS resolution
        # NOTE: May not work unless the user configures their default mainnet provider.
        ecosystem = self.network_manager.get_ecosystem("ethereum")
        mainnet = ecosystem.get_network("mainnet")
        if not (provider := mainnet.default_provider):
            return None

        elif not isinstance(provider, Web3Provider):
            logger.warning(
                "Unable to connect to mainnet provider to "
                "perform ENS lookup (must be a Web3Provider)"
            )
            return None

        if not provider.is_connected:
            # Connect if needed.
            try:
                provider.connect()
            except ProviderError:
                logger.warning(
                    "Unable to connect to mainnet provider to perform ENS lookup. "
                    "Try changing your default mainnet provider."
                )
                return None

        return provider

    def is_convertible(self, value: Any) -> bool:
        if not isinstance(value, str):
            return False

        elif "." not in value:
            return False

        elif not ENS.is_valid_name(value):
            return False

        elif value in self.address_cache:
            return True

        else:
            try:
                provider = self.mainnet_provider
                if not provider:
                    return False

                elif not (ens := provider.web3.ens):
                    return False

                try:
                    address = ens.address(value)
                except CannotHandleRequest:
                    # Either this is not actually mainnet or our head is
                    # pointed before ENS existed.
                    return False

                if address is not None:
                    self.address_cache[value] = address
                    return True

                return False

            except (NetworkError, ProviderError):
                return False

    def convert(self, value: str) -> AddressType:
        if value in self.address_cache:
            return self.address_cache[value]

        elif not (provider := self.mainnet_provider):
            # Should never get here.
            raise ValueError(f"Unable to convert ENS value '{value}'.")

        # TODO: Switch to using ENS SDK
        if not (ens := provider.web3.ens):
            # Should never get here.
            raise ValueError(f"Unable to convert ENS value '{value}'.")

        address = ens.address(value)
        self.address_cache[value] = address
        return address



import codecs
import functools

from eth_utils import is_bytes, keccak
from hexbytes import HexBytes

try:
    from ens.utils import raw_name_to_hash  # type: ignore
except ImportError:
    raw_name_to_hash = None

# https://github.com/ethereum/EIPs/blob/master/EIPS/eip-137.md#namehash-algorithm


def _combine(f, g):
    return lambda x: f(g(x))


def _compose(*functions):
    return functools.reduce(_combine, functions, lambda x: x)


def _sub_hash(value, label):
    return keccak(value + keccak(label))


def _manual_namehash(name: str, encoding=None) -> HexBytes:
    node = b"\x00" * 32
    if name:
        if encoding is None:
            if is_bytes(name):
                encoded_name = name
            else:
                encoded_name = codecs.encode(name, "utf8")  # type: ignore
        else:
            encoded_name = codecs.encode(name, encoding)

        labels = encoded_name.split(b".")  # type: ignore

        return HexBytes(
            _compose(*(functools.partial(_sub_hash, label=label) for label in labels))(  # noqa: 501
                node
            )
        )
    return HexBytes(node)


def namehash(name: str) -> HexBytes:
    namehash_func = raw_name_to_hash if raw_name_to_hash else _manual_namehash
    return namehash_func(name)



import ape
from web3.exceptions import CannotHandleRequest

from tests.conftest import negative_tests


def test_is_convertible(converter):
    assert converter.is_convertible("test.eth")


@negative_tests
def test_is_not_convertible(converter, value):
    assert not converter.is_convertible(value)


def test_is_not_convertible_when_ens_address_call_fails(converter, mainnet_provider):
    """
    Tests against a bug where this would fail in an uncaught way rather
    than saying ``is_convertible=False``.
    """
    mainnet_provider.web3.ens.address.side_effect = CannotHandleRequest
    # NOTE: Using test2.eth since it should be free from the cache (not used elsewhere)
    assert not converter.is_convertible("test2.eth")


def test_address_cache(converter, address):
    converter.address_cache["test.eth"] = address
    assert converter.convert("test.eth") == address


def test_mainnet_fork(converter, mocker):
    get_eco_spy = mocker.spy(converter.network_manager, "get_ecosystem")

    with ape.networks.ethereum.mainnet_fork.use_provider("mock-mainnet-fork"):
        get_eco_spy.reset_mock()
        converter.convert("test.eth")

    # Should not have to reconnect
    assert not get_eco_spy.call_count


def test_other_ecosystem_mainnet(converter, mocker):
    get_eco_spy = mocker.spy(converter.network_manager, "get_ecosystem")

    with ape.networks.polygon.mainnet.use_provider("mock-polygon-mainnet"):
        get_eco_spy.reset_mock()
        converter.convert("test.eth")

    # It has to re-connect to Ethereum mainnet temporarily
    get_eco_spy.assert_called_once_with("ethereum")
